#ifndef COMPRESSER_RLE_H
#define COMPRESSER_RLE_H

#include "rgb_value.h"
#include <cstdint>
#include <vector>
#include <cassert>
#include <limits>
#include <type_traits>

//
// Compresse RLE un vecteur source. On pourrait mieux faire encore en
// procdant  partir d'une squence d'itrateurs plutt qu' partir
// d'un conteneur spcifique, mais ce sera correct tel quel pour nos
// fins ici.
//
// Il est possible que nous devions morceler le traitement et que
// cette fonction devienne interruptible puisque le temps de traitement
// sur de longues squences peut crotre de manire arbitraire
//
// Si le vecteur de destination n'est pas implment par COW, il est
// possible que le return soit coteux. Dans ce cas, modifier la
// smantique de la fonction peut tre une approche  considrer (passer
// le vecteur de destination par rfrence ou oprer sur une squence
// de destination prise sous forme d'itrateurs et de back_inserter)
//
// L'opration push_back() est subtile car elle est en temps constant
// amorti, ce qui poeut causer des bursts occasionnels de temps processeur
// ici. Il est possible que, mme si cela dgrade la performance d'ensemble,
// on doive demander  l'appelant de prallouer de la mmoire pour la
// destination, ou mme que la destination soit un tableau de taille fixe
// pour que l'insertion soit vritablement en temps constant.
//
// Rgles:
// - Src peut tre converti naturellement en Dest
// - une diffrence d'itrateurs peut tre convertie naturellement en Dest
// - ceci fait partie de la nature-mme d'une compression RLE
//
// RAFFINEMENTS APPORTS:
//
// - restructurer le code pour qu'il soit exprim sous forme d'oprations
//   sur des itrateurs, pour faciliter la reprise en un point arbitraire;
// - introduire la question du non dpassement d'un seuil de temps;
// - distinguer les ides de interrompu et de complt
//
// RGLES AJUSTES:
//
// - l'appelant doit s'assurer du non dbordement de la destination, en
//   supplant une destination assez grande en tout temps ou en offrant
//   un adaptateur de croissance (p.ex.: un back_inserter)
//

#include <iterator>

//
// Note technique: le type dest_type doit tre fourni car, pour un Itt
// dfinie comme adaptateur d'itrateur comme un back_insert_iterator,
// pour une raison que j'ignore (je pense la comprendre, pour tre bien
// honnte, mais c'est un hunch), iterator_traits<Itt>::value_type est
// void. Il aurait t plus lgant pour cet algorithme de pouvoir tirer
//  la compilation dest_type du type DestIt... Enfin.
//
template <class dest_type, class DestIt, class SrcIt, class Pred>
   std::pair<SrcIt, DestIt>
      compresser_rle(SrcIt debut, SrcIt fin, DestIt dest, Pred pred) {
      using namespace std;
      using src_type = typename iterator_traits<SrcIt>::value_type;
      static_assert(is_convertible_v<src_type, dest_type>);
      //
      // Squence vide ou contraintes trop serres (IMPORTANT: il faut
      // retourner un tat comprenant debut, pas fin, au cas o l'arrt
      // serait signal par le prdicat!)
      //
      if (debut == fin || !pred())
         return { debut, dest };

      //
      // Pour faciliter la logique, nous testerons la condition d'arrt
      // seulement une fois une squence RLE conclue
      //
      src_type tgt = *debut;
      auto pos = next(debut);
      while (pos != fin) {
         //
         // Consommer une squence RLE
         //
         auto dist = static_cast<dest_type>(distance(debut, pos)); // ceci peut tre coteux!!!
         if (*pos != tgt || dist == numeric_limits<dest_type>::max()) {
            *dest++ = dist;
            *dest++ = tgt;
            tgt = *pos;
            debut = pos;
            if (!pred())
               return { debut, dest };
         }
         ++pos;
      }
      if (auto dist = static_cast<dest_type>(distance(debut, pos)); dist) {
         *dest++ = dist;
         *dest++ = tgt;
      }
      return { fin, dest };
   }

#endif