#ifndef RGB_VALUE_H
#define RGB_VALUE_H

#include <cstdint>
#include <algorithm>
#include <vector>
#include <iterator>
#include <cassert>
#include <type_traits>

//
// Deux implmentations d'un triple RGB pour reprsenter les 
// bitmaps Windows v3. Ceci constitue un seuil smantique plus
// lev que celui des octets bruts. J'ai essay deux versions,
// l'une compacte et l'autre non, et la version compacte tait
// lgrement plus rapide que la version non compacte sur ma
// plateforme personnelle. Si la chose diffre chez vous, il y
// a une directive using, plus bas, qui permet de passer de
// l'une  l'autre. Les interfaces sont identiques.
//
// La capacit de passer de rgb_value  int_32 implicitement
// permet d'implmenter la compression RLE en termes de squences
// RGB, ce qui donne une compression nettement meilleure qu'elle
// ne le serait sur des octets pour la majorit des bitmaps.
//
namespace packed_rgb {
   class rgb_value {
   public:
      using byte = std::uint8_t;
   private:
      enum { ROUGE, VERT, BLEU, ALPHA, NBYTES };
      byte values[NBYTES]{}; // R G B (A)
   public:
      rgb_value() = default;
      rgb_value(byte r, byte g, byte b) noexcept {
         red(r);
         green(g);
         blue(b);
         alpha({});
      }
      void red(byte val) noexcept {
         values[ROUGE] = val;
      }
      byte red() const noexcept {
         return values[ROUGE];
      }
      void green(byte val) noexcept {
         values[VERT] = val;
      }
      byte green() const noexcept {
         return values[VERT];
      }
      void blue(byte val) noexcept {
         values[BLEU] = val;
      }
      byte blue() const noexcept {
         return values[BLEU];
      }
      void alpha(byte val) noexcept {
         values[ALPHA] = val;
      }
      byte alpha() const noexcept {
         return values[ALPHA];
      }
      operator std::uint32_t() const noexcept {
         return *reinterpret_cast<const std::uint32_t*>(values);
      }
      bool operator==(const rgb_value &autre) const noexcept {
         return static_cast<std::uint32_t>(*this) == static_cast<std::uint32_t>(autre);
      }
      bool operator!=(const rgb_value &autre) const noexcept {
         return !(*this == autre);
      }
   };
}

namespace not_packed_rgb {
   class rgb_value {
   public:
      using byte = std::uint8_t;
   private:
      byte rouge, vert, bleu;
   public:
      rgb_value(byte r, byte g, byte b) noexcept {
         red(r);
         green(g);
         blue(b);
         alpha({});
      }
      void red(byte val) noexcept { rouge = val; }
      byte red() const noexcept { return rouge; }
      void green(byte val) noexcept { vert = val; }
      byte green() const noexcept { return vert; }
      void blue(byte val) noexcept { bleu = val; }
      byte blue() const noexcept { return bleu; }
      void alpha(byte) noexcept {  }
      byte alpha() const noexcept { return {}; }
      operator std::uint32_t() const noexcept {
         return std::uint32_t((red() << 24) | (green() << 16) | (blue() << 8));
      }
      bool operator==(const rgb_value &autre) const noexcept {
         return red() == autre.red() && green() == autre.green() && blue() == autre.blue();
      }
      bool operator!=(const rgb_value &autre) const noexcept {
         return !(*this == autre);
      }
   };
}

using packed_rgb::rgb_value; // lgrement plus rapide que not_packed_rgb::rgb_value

//
// Conversion d'une squence d'octets en squence. Cette opration est
// plus coteuse qu'il n'y parat, alors il est possible qu'elle doive
// tre rendue interruptible
//
template <class It>
   std::vector<rgb_value> to_rgb_values(It debut, It fin) {
      using namespace std;
      static_assert (is_same <iterator_traits<It>::value_type, std::uint8_t>::value);
      assert(distance(debut, fin) % 3 == 0);
      // ---------------------------------------------
      // Test qui donne de mauvais rsultats
      //
      //rgb_value _default(0,0,0);
      //vector<rgb_value> v(distance(debut, fin) / 3, _default);
      // ---------------------------------------------
      vector<rgb_value> v;
      for (; debut != fin; debut += 3)
         v.emplace_back(*(debut + 2), *(debut + 1), *debut);
      return v;
   }

#endif
